import os
import sys
import signal
from time import sleep
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor

import pytest
from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.ErrorHandler import ErrorHandler
from pypushflow.stop_workflows import DEFAULT_STOP_SIGNALS


@pytest.mark.parametrize("forced_interruption", [True, False])
@pytest.mark.parametrize("sleep_time", [0, 2])
def test_stop_workflow(forced_interruption, sleep_time, workflow_cleanup):
    if os.name == "nt" and forced_interruption:
        pytest.skip("not supported on windows")
    testWorkflow1 = WorkflowSleep("Test workflow Sleep")

    def run_normal():
        inData = {"sleep_time": 0, "counter": 0}
        future = executor.submit(
            testWorkflow1.run,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
        )
        result = future.result()
        assert "WorkflowException" not in result
        assert result["counter"] == 3

    def run_stopped():
        inData = {"sleep_time": 2, "counter": 0}
        future = executor.submit(
            testWorkflow1.run,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
        )
        if sleep_time:
            sleep(sleep_time)
        testWorkflow1.stop(
            reason="reason for stopping", forced_interruption=forced_interruption
        )
        result = future.result()
        assert result["counter"] < 3
        assert "WorkflowException" in result

    with ThreadPoolExecutor() as executor:
        run_normal()
        run_stopped()
        run_stopped()
        run_normal()


@pytest.mark.parametrize("forced_interruption", [True, False])
@pytest.mark.parametrize("sleep_time", [2])
def test_stop_signal_workflow(
    forced_interruption, sleep_time, skip_when_gevent, skip_on_windows
):
    def run_normal():
        inData = {"sleep_time": 0, "counter": 0}
        future = executor.submit(
            _run_workflow,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
            forced_interruption=forced_interruption,
        )
        result = future.result()
        assert "WorkflowException" not in result
        assert result["counter"] == 3

    def run_stopped():
        inData = {"sleep_time": 2, "counter": 0}
        future = executor.submit(
            _run_workflow,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
            forced_interruption=forced_interruption,
        )
        if sleep_time:
            sleep(sleep_time)
        for pid in list(executor._processes):
            # TODO: doesn't work on windows (it kills the subprocess)
            os.kill(pid, DEFAULT_STOP_SIGNALS[0])

        result = future.result()
        assert result["counter"] < 3
        assert "WorkflowException" in result

    if sys.version_info < (3, 7):
        pool = ProcessPoolExecutor(max_workers=1)
        pool.submit(_process_initializer).result()
    else:
        pool = ProcessPoolExecutor(initializer=_process_initializer)

    with pool as executor:
        run_normal()
        run_stopped()
        run_stopped()
        run_normal()


def _process_initializer():
    try:
        signal.signal(DEFAULT_STOP_SIGNALS[0], signal.SIG_IGN)
    except (OSError, AttributeError, ValueError, RuntimeError):
        pass


def _run_workflow(*args, forced_interruption: bool = False, **kwargs):
    testWorkflow1 = WorkflowSleep(
        "Test workflow Sleep",
        stop_on_signals=True,
        forced_interruption=forced_interruption,
    )
    return testWorkflow1.run(*args, **kwargs)


class WorkflowSleep(Workflow):
    def __init__(self, name, **kw):
        super().__init__(name, **kw)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.errorActor = ErrorHandler(parent=self, thread_counter=ctr)
        self.pythonActor1 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.pythonActor2 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.pythonActor3 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.pythonActor1)
        self.pythonActor1.connect(self.pythonActor2)
        self.pythonActor2.connect(self.pythonActor3)
        self.pythonActor3.connect(self.stopActor)
        self.errorActor.connect(self.stopActor)
        self.pythonActor1.connectOnError(self.errorActor)
        self.pythonActor2.connectOnError(self.errorActor)
        self.pythonActor3.connectOnError(self.errorActor)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "07/04/2021"

from pypushflow.Workflow import Workflow
from pypushflow.Submodel import Submodel
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Submodel6(Submodel):
    """
    Submodel containing one python actor which has a long execution time
    """

    def __init__(self, parent, name, thread_counter):
        super().__init__(parent=parent, name=name, thread_counter=thread_counter)
        self.pythonActorAdd = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Submodel Actor Add",
            errorHandler=self,
            thread_counter=thread_counter,
        )
        self.getPort("In").connect(self.pythonActorAdd)
        self.pythonActorAdd.connect(self.getPort("Out"))


class Workflow6(Workflow):
    """
    Workflow containing one start actor,
    one submodel which has a long execution and one stop actor with short timeout.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(self, thread_counter=ctr)
        self.pythonActorAdd1 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Add Actor 1",
            thread_counter=ctr,
        )
        self.pythonActorAdd2 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Add Actor 2",
            thread_counter=ctr,
        )
        self.submodel6 = Submodel6(self, name="Submodel 6", thread_counter=ctr)
        self.stopActor = StopActor(self, thread_counter=ctr)
        self.startActor.connect(self.pythonActorAdd1)
        self.pythonActorAdd1.connect(self.submodel6.getPort("In"))
        self.submodel6.getPort("Out").connect(self.pythonActorAdd2)
        self.pythonActorAdd2.connect(self.stopActor)


class TestWorkflow6(WorkflowTestCase):
    def test_workflow6(self):
        workflow6 = Workflow6("Test workflow 6")
        inData = {"value": 1}
        outData = workflow6.run(
            inData, timeout=5, scaling_workers=False, max_workers=-1
        )
        self.assertEqual(outData["value"], 4)



from multiprocessing.pool import RemoteTraceback as MpRemoteTraceback

try:
    from .billiard import RemoteTraceback as BRemoteTraceback
except ImportError:
    BRemoteTraceback = None


def extract_remote_exception(exception: BaseException) -> BaseException:
    if isinstance(exception.__cause__, MpRemoteTraceback):
        return exception.__cause__
    if BRemoteTraceback is not None and isinstance(
        exception.__cause__, BRemoteTraceback
    ):
        return exception.__cause__
    return exception
